"""
Configure a CashFlow class to handle calculations of cashflow indicators.
"""

from dataclasses import dataclass, field
from datetime import date
import numpy as np

from pyscnomics.econ.selection import FluidType
from pyscnomics.econ import indicator


class CashFlowException(Exception):
    """Exception to raise for a misuse of CashFlow class"""

    pass


@dataclass
class CashFlow:
    """
    A class representing a cash flow with various financial indicators.

    Parameters
    ----------
    start_date : date
        The start date of the cash flow.
    end_date : date
        The end date of the cash flow.
    cash : np.ndarray
        An array representing cash flows over time.
    cashed_year : np.ndarray
        An array representing the cashed year of a cashflow.
    cash_allocation : FluidType, optional
        The cash allocation type for cash flows.
        Default is FluidType.ALL.
    """

    start_date: date
    end_date: date
    cash: np.ndarray
    cashed_year: np.ndarray
    cash_allocation: FluidType = field(default=FluidType.ALL)

    # Attributes to be defined later
    project_duration: int = field(default=None, init=False, repr=False)
    project_years: np.ndarray = field(default=None, init=False, repr=False)

    def __post_init__(self):
        # Specify project duration and project years, raise error for inappropriate start date
        if self.start_date <= self.end_date:
            self.project_duration = self.end_date.year - self.start_date.year + 1
            self.project_years = np.arange(
                self.start_date.year, self.end_date.year + 1, 1
            )

        else:
            raise CashFlowException(
                f"start date {self.start_date} "
                f"is after the end date: {self.end_date}"
            )

        # Check input data for unequal length of arrays
        if len(self.cash) != len(self.cashed_year):
            raise CashFlowException(
                f"Unequal length of array: "
                f"cash: {len(self.cash)}, "
                f"cash_year: {len(self.cashed_year)}, "
            )

        # Raise an error message: cashed_year is after the end year of the project
        if np.max(self.cashed_year) > self.end_date.year:
            raise CashFlowException(
                f"Cashed year ({np.max(self.cashed_year)}) "
                f"is after the end year of the project ({self.end_date.year})"
            )

        # Raise an error message: cashed_year is before the start year of the project
        if np.min(self.cashed_year) < self.start_date.year:
            raise CashFlowException(
                f"Cashed year ({np.min(self.cashed_year)}) "
                f"is before the start year of the project ({self.start_date.year})"
            )

    def __len__(self):
        return self.project_duration

    def __eq__(self, other):
        # Between two instances of CashFlow
        if isinstance(other, CashFlow):
            return all(
                (
                    self.start_date == other.start_date,
                    self.end_date == other.end_date,
                    self.cash_allocation == other.cash_allocation,
                    np.allclose(self.cash, other.cash),
                    np.allclose(self.cashed_year, other.cashed_year),
                )
            )

        # Between an instance of CashFlow and an integer/float
        elif isinstance(other, (int, float)):
            return np.allclose(sum(self.cash), other)

        else:
            return False

    def __lt__(self, other):
        # Between two instances of CashFlow
        if isinstance(other, CashFlow):
            return np.sum(self.cash) < np.sum(other.cash)

        # Between an instance of CashFlow and an integer/float
        elif isinstance(other, (int, float)):
            return np.sum(self.cash) < other

        else:
            raise CashFlowException(
                f"Must compare an instance of CashFlow with another instance of "
                f"CashFlow, an integer, or a float. "
            )

    def __le__(self, other):
        # Between two instances of CashFlow
        if isinstance(other, CashFlow):
            return np.sum(self.cash) <= np.sum(other.cash)

        # Between an instance of CashFlow and an integer/float
        elif isinstance(other, (int, float)):
            return np.sum(self.cash) <= other

        else:
            raise CashFlowException(
                f"Must compare an instance of CashFlow with another instance of "
                f"CashFlow, an integer, or a float. "
            )

    def __gt__(self, other):
        # Between two instances of CashFlow
        if isinstance(other, CashFlow):
            return np.sum(self.cash) > np.sum(other.cash)

        # Between an instance of CashFlow and an integer/float
        elif isinstance(other, (int, float)):
            return np.sum(self.cash) > other

        else:
            raise CashFlowException(
                f"Must compare an instance of CashFlow with another instance of "
                f"CashFlow, an integer, or a float."
            )

    def __ge__(self, other):
        # Between two instances of CashFlow
        if isinstance(other, CashFlow):
            return np.sum(self.cash) >= np.sum(other.cash)

        # Between an instance of CashFlow and an integer/float
        elif isinstance(other, (int, float)):
            return np.sum(self.cash) >= other

        else:
            raise CashFlowException(
                f"Must compare an instance of CashFlow with another instance of "
                f"CashFlow, an integer, or a float."
            )

    def __add__(self, other):
        # Between an instance of CashFlow and another instance of CashFlow
        if isinstance(other, CashFlow):
            start_year = min(self.start_date.year, other.start_date.year)
            start_month = min(self.start_date.month, other.start_date.month)
            start_day = min(self.start_date.day, other.start_date.day)

            end_year = max(self.end_date.year, other.end_date.year)
            end_month = max(self.end_date.month, other.end_date.month)
            end_day = max(self.end_date.day, other.end_date.day)

            return CashFlow(
                start_date=date(year=start_year, month=start_month, day=start_day),
                end_date=date(year=end_year, month=end_month, day=end_day),
                cash=np.concatenate((self.cash, other.cash)),
                cashed_year=np.concatenate((self.cashed_year, other.cashed_year)),
            )

        # Between an instance of CashFlow and an integer/float
        elif isinstance(other, (int, float)):
            return self.allocate() + other

        else:
            raise CashFlowException(
                f"Must add between an instance of CashFlow "
                f"with another instance of CashFlow or an integer/float. "
                f"{other}({other.__class__.__qualname__}) is not "
                f"an instance of Cashflow nor an integer/float."
            )

    def __sub__(self, other):
        # Between an instance of CashFlow and another instance of CashFlow
        if isinstance(other, CashFlow):
            start_year = min(self.start_date.year, other.start_date.year)
            start_month = min(self.start_date.month, other.start_date.month)
            start_day = min(self.start_date.day, other.start_date.day)

            end_year = max(self.end_date.year, other.end_date.year)
            end_month = max(self.end_date.month, other.end_date.month)
            end_day = max(self.end_date.day, other.end_date.day)

            return CashFlow(
                start_date=date(year=start_year, month=start_month, day=start_day),
                end_date=date(year=end_year, month=end_month, day=end_day),
                cash=np.concatenate((self.cash, -other.cash)),
                cashed_year=np.concatenate((self.cashed_year, other.cashed_year)),
            )

        # Between an instance of CashFlow and an integer/float
        elif isinstance(other, (int, float)):
            return self.allocate() - other

        else:
            raise CashFlowException(
                f"Must subtract between an instance of CashFlow "
                f"with another instance of CashFlow or an integer/float. "
                f"{other}({other.__class__.__qualname__}) is not "
                f"an instance of Cashflow nor an integer/float."
            )

    def __rsub__(self, other):
        return self.__sub__(other)

    def __mul__(self, other):
        # Multiplication is allowed only with an integer/float
        if isinstance(other, (int, float)):
            return CashFlow(
                start_date=self.start_date,
                end_date=self.end_date,
                cash=self.cash * other,
                cashed_year=self.cashed_year,
                cash_allocation=self.cash_allocation,
            )

        else:
            raise CashFlowException(
                f"Must multiply with an integer or a float; "
                f"{other}({other.__class__.__qualname__}) is not an integer nor a float."
            )

    def __rmul__(self, other):
        return self.__mul__(other)

    def __truediv__(self, other):
        # Between an instance of CashFlow and another instance of CashFlow
        if isinstance(other, CashFlow):
            return np.sum(self.cash) / np.sum(other.cash)

        # Between an instance of CashFlow and an integer/float
        elif isinstance(other, (int, float)):
            # Cannot divide by zero
            if other == 0:
                raise CashFlowException(f"Cannot divide by zero")

            else:
                return CashFlow(
                    start_date=self.start_date,
                    end_date=self.end_date,
                    cash=self.cash / other,
                    cashed_year=self.cashed_year,
                    cash_allocation=self.cash_allocation,
                )

        else:
            raise CashFlowException(
                f"Must divide with an instance of CashFlow, integer or a float; "
                f"{other}({other.__class__.__qualname__}) is not an instance "
                f"of CashFlow nor an integer/float."
            )

    def allocate(self) -> np.ndarray:
        """
        Allocates cash flows over the project duration.

        This method calculates the allocation of cash flows over the project duration
        based on the cash, cashed_year, and project start_year data.

        Returns
        -------
        cash_allocation: np.ndarray
            The cashflow in which each of the associated elements is aligned with
            its corresponding cashed year.

        Notes
        -----
        The calculation procedures are as follows:
        (1) Generate an array depicting the difference between the cashed_year and the start_year
            of the project,
        (2) Calculate the number of occurrence of each of the elements induced from (1)
            using method numpy.bincount(),
        (3) Multiply the results of (2) with the corresponding cash data to obtain an array
            of cash that is aligned with the corresponding year. To do so, we use the argument
            'weights' of numpy.bincount() method.
        (4) If the length of cost_allocation does not equal to the project duration, modify
            cash_allocation array by adding zeros to the last elements as many as the difference
            between the initial length of cash_allocation and the project duration.
        """
        cash_allocation = np.bincount(
            self.cashed_year - self.start_date.year, weights=self.cash
        )
        zeros = np.zeros(self.project_duration - len(cash_allocation))

        return np.concatenate((cash_allocation, zeros))

    def npv(self, disc_rate: float = 0.1):
        """
        Calculate the net present value (NPV) of the cash flows.

        Parameters
        ----------
        disc_rate : float, optional
            The discount rate used in financial calculations. Default is 0.1 (10%).

        Returns
        -------
        float
            The calculated NPV value.
        """
        return indicator.npv(cashflow=self.cash, disc_rate=disc_rate)

    def xnpv(self, disc_rate: float = 0.1):
        """
        Calculate the NPV of cash flows at irregular intervals.

        Parameters
        ----------
        disc_rate : float, optional
            The discount rate used in financial calculations. Default is 0.1 (10%).

        Returns
        -------
        float
            The calculated NPV value at irregular intervals.
        """
        return indicator.xnpv(
            cashflow=self.cash,
            start_date=self.start_date,
            end_date=self.end_date,
            disc_rate=disc_rate,
        )

    def irr(self):
        """
        Calculate the internal rate of return (IRR) of the cash flows.

        Returns
        -------
        float
            The calculated IRR value.
        """
        return indicator.irr(cashflow=self.cash)

    def xirr(self):
        """
        Calculate the IRR of cash flows at irregular intervals.

        Returns
        -------
        float
            The calculated IRR value at irregular intervals.
        """
        return indicator.xirr(
            cashflow=self.cash,
            start_date=self.start_date,
            end_date=self.end_date,
        )

    def pot(self):
        """
        Calculate the potential investment return of the cash flows.
        (pay out time)

        Returns
        -------
        float
            The calculated potential investment return value.
        """
        return indicator.pot(cashflow=self.cash)



"""
Collection of functions to administer Monte Carlo simulation.
"""

import os as os
import numpy as np
import pandas as pd
import multiprocessing as mp
from scipy.stats import uniform, triang, truncnorm

from pyscnomics.tools.summary import get_summary
from pyscnomics.io.aggregator import Aggregate
from pyscnomics.optimize.adjuster import AdjustData

from pyscnomics.contracts.project import BaseProject
from pyscnomics.contracts.costrecovery import CostRecovery
from pyscnomics.contracts.grossplit import GrossSplit
from pyscnomics.contracts.transition import Transition


class MonteCarloException(Exception):
    """ Exception to be raised for a misuse of MonteCarlo class """

    pass


def get_montecarlo_data(data: Aggregate) -> dict:
    """
    Prepares Monte Carlo simulation data from the given Aggregate object.

    Parameters
    ----------
    data: Aggregate
        The input data containing Monte Carlo simulation data.

    Returns
    -------
    dict
        A dictionary containing the prepared Monte Carlo simulation data.

    Raises
    ------
    MonteCarloException
        If any of the required data is missing or inconsistent.
    """
    # Prepare a container to store all the necessary data
    uncertainty_data = {}

    # Prepare attribute run_number
    if pd.isna(data.montecarlo_data.run_number):
        raise MonteCarloException(
            f"The data for run number is missing. Please check the input data."
        )

    uncertainty_data["run_number"] = int(data.montecarlo_data.run_number)

    # Prepare attribute parameter
    if not isinstance(data.montecarlo_data.parameter, list):
        raise MonteCarloException(
            f"Attribute parameter must be provided in a list datatype, "
            f"not {data.montecarlo_data.parameter.__class__.__qualname__}"
        )

    uncertainty_data["parameter"] = data.montecarlo_data.parameter

    # Prepare attribute distribution
    if not isinstance(data.montecarlo_data.distribution, list):
        raise MonteCarloException(
            f"Attribute distribution must be provided in a list datatype, "
            f"not {data.montecarlo_data.distribution.__class__.__qualname__}"
        )

    uncertainty_data["distribution"] = data.montecarlo_data.distribution

    # Prepare attribute min_values
    if not isinstance(data.montecarlo_data.min_values, np.ndarray):
        raise MonteCarloException(
            f"Attribute min_values must be provided in a numpy ndarray datatype, "
            f"not {data.montecarlo_data.min_values.__class__.__qualname__}"
        )

    min_values_nan = list(filter(lambda i: pd.isna(i), data.montecarlo_data.min_values))
    if len(min_values_nan) > 0:
        raise MonteCarloException(
            f"Minimum values data are incomplete. Please check the input data."
        )

    uncertainty_data["min_values"] = data.montecarlo_data.min_values.astype(np.float_)

    # Prepare attribute mean_values
    if not isinstance(data.montecarlo_data.mean_values, np.ndarray):
        raise MonteCarloException(
            f"Attribute mean_values must be provided in a numpy ndarray datatype, "
            f"not {data.montecarlo_data.mean_values.__class__.__qualname__}"
        )

    mean_values_nan = list(filter(lambda i: pd.isna(i), data.montecarlo_data.mean_values))
    if len(mean_values_nan) > 0:
        raise MonteCarloException(
            f"Mean values data are incomplete. Please check the input data."
        )

    uncertainty_data["mean_values"] = data.montecarlo_data.mean_values.astype(np.float_)

    # Prepare attribute max_values
    if not isinstance(data.montecarlo_data.max_values, np.ndarray):
        raise MonteCarloException(
            f"Attribute max_values must be provided in a numpy ndarray datatype, "
            f"not {data.montecarlo_data.max_values.__class__.__qualname__}"
        )

    max_values_nan = list(filter(lambda i: pd.isna(i), data.montecarlo_data.max_values))
    if len(max_values_nan) > 0:
        raise MonteCarloException(
            f"Maximum values data are incomplete. Please check the input data."
        )

    uncertainty_data["max_values"] = data.montecarlo_data.max_values.astype(np.float_)

    # Prepare attribute standard deviation
    if not isinstance(data.montecarlo_data.std_dev, np.ndarray):
        raise MonteCarloException(
            f"Attribute std_dev must be provided in a numpy ndarray datatype, "
            f"not {data.montecarlo_data.std_dev.__class__.__qualname__}"
        )

    for i, val in enumerate(uncertainty_data["distribution"]):
        if val == "Uniform" or val == "Triangular":
            data.montecarlo_data.std_dev[i] = 0.0

    std_dev_nan = list(filter(lambda i: pd.isna(i), data.montecarlo_data.std_dev))
    if len(std_dev_nan) > 0:
        raise MonteCarloException(
            f"Std deviation data are incomplete. Please check the input data."
        )

    uncertainty_data["std_dev"] = data.montecarlo_data.std_dev.astype(np.float_)

    # Throw exception for inappropriate conditions
    for i, val in enumerate(uncertainty_data["min_values"]):
        if (
            uncertainty_data["min_values"][i] > uncertainty_data["max_values"][i]
            or uncertainty_data["min_values"][i] > uncertainty_data["mean_values"][i]
            or uncertainty_data["mean_values"][i] > uncertainty_data["max_values"][i]
        ):
            raise MonteCarloException(
                f"Inappropriate minimum, mean, or maximum value data. "
                f"Please check the input data."
            )

    # Identify zero mean values
    mean_values_not_zero = np.argwhere(uncertainty_data["mean_values"] != 0.0).ravel()

    return uncertainty_data, mean_values_not_zero


def get_multipliers_montecarlo(
    run_number: int,
    distribution: str,
    min_value: float,
    mean_value: float,
    max_value: float,
    std_dev: float,
) -> np.ndarray:
    """
    Generate an array of multipliers for Monte Carlo simulation based on the specified distribution.

    Parameters
    ----------
    run_number: int
        Number of runs.
    distribution: str
        Type of distribution ("Uniform", "Triangular", or "Normal").
    min_value: float
        Minimum value for the distribution.
    mean_value: float
        Mean (or central value) for the distribution.
    max_value: float
        Maximum value for the distribution.
    std_dev: float
        Standard deviation for the normal distribution.

    Returns
    -------
    multipliers: np.ndarray
        Array of multipliers generated using Monte Carlo simulation.

    Notes
    -----
    - For "Uniform" distribution, the function uses a uniform random variable.
    - For "Triangular" distribution, the function uses a triangular random variable.
    - For "Normal" distribution, the function uses a truncated normal random variable.
    """
    # Uniform distribution
    if distribution == "Uniform":
        # Modify minimum and maximum values
        params = [min_value, max_value]
        attrs_updated = {
            key: float(params[i] / mean_value)
            for i, key in enumerate(["min_value", "max_value"])
        }

        # Determine multipliers
        multipliers = uniform.rvs(
            loc=attrs_updated["min_value"],
            scale=attrs_updated["max_value"] - attrs_updated["min_value"],
            size=run_number,
        )

    # Triangular distribution
    elif distribution == "Triangular":
        # Modify minimum and maximum values
        params = [min_value, mean_value, max_value]
        attrs_updated = {
            key: float(params[i] / mean_value)
            for i, key in enumerate(["min_value", "mean_value", "max_value"])
        }

        # Determine mode (central point)
        c = (
            (attrs_updated["mean_value"] - attrs_updated["min_value"]) /
            (attrs_updated["max_value"] - attrs_updated["min_value"])
        )

        # Determine multipliers
        multipliers = triang.rvs(
            c=c,
            loc=attrs_updated["min_value"],
            scale=attrs_updated["max_value"] - attrs_updated["min_value"],
            size=run_number,
        )

    # Normal distribution
    elif distribution == "Normal":
        # Modify minimum and maximum values
        params = [min_value, mean_value, max_value, std_dev]
        attrs_updated = {
            key: float(params[i] / mean_value)
            for i, key in enumerate(["min_value", "mean_value", "max_value", "std_dev"])
        }

        # Determine z-values
        zvalues = {
            key: float(
                (attrs_updated[key] - attrs_updated["mean_value"]) / attrs_updated["std_dev"]
            )
            for key in ["min_value", "mean_value", "max_value"]
        }

        # Determine multipliers
        multipliers_init = truncnorm.rvs(
            a=zvalues["min_value"],
            b=zvalues["max_value"],
            loc=zvalues["mean_value"],
            scale=1,
            size=run_number,
        )

        multipliers = (multipliers_init * attrs_updated["std_dev"]) + attrs_updated["mean_value"]

    else:
        raise MonteCarloException(
            f"The type of distribution is unavailable. "
            f"Please select type of distribution between: "
            f"(i) Uniform, (ii) Triangular, or (iii) Normal."
        )

    return multipliers


def run_montecarlo(
    contract: BaseProject | CostRecovery | GrossSplit | Transition,
    contract_arguments: dict,
    summary_arguments: dict,
) -> tuple:
    """
    Function to run Monte Carlo simulation.

    Parameters
    ----------
    contract: BaseProject | CostRecovery | GrossSplit | Transition
        The contract type for the project.
    contract_arguments: dict
        Arguments required for running the specified contract.
    summary_arguments: dict
        Arguments for generating the project summary.

    Returns
    -------
    tuple:
        A tuple containing the following project metrics:
        - NPV (Net Present Value)
        - IRR (Internal Rate of Return)
        - PI (Profitability Index)
        - POT (Pay Out Time)
        - Gov Take (Government Take)
        - Net Share (Contractor's Net Share)
    """
    contract.run(**contract_arguments)
    contract_summary = get_summary(**summary_arguments)

    return (
        contract_summary["ctr_npv"],
        contract_summary["ctr_irr"],
        contract_summary["ctr_pi"],
        contract_summary["ctr_pot"],
        contract_summary["gov_take"],
        contract_summary["ctr_net_share"],
    )


def execute_montecarlo(
    data: Aggregate,
    workbook_path: str,
    uncertainty_data: dict,
    target: list,
    mult: np.ndarray,
    threshold_parallel: int = 500,
) -> np.ndarray:
    """
    A target function to run MonteCarlo simulation.

    Parameters
    ----------
    data: Aggregate
        The aggregate data for the simulation.
    workbook_path: str
        The path to the workbook containing data.
    uncertainty_data: dict
        Data associated with uncertainty analysis.
    target: list
        A list of targets for the simulation.
    mult: np.ndarray
        A numpy array containing multipliers.
    threshold_parallel: int
        A threshold value to activate parallel computation.

    Returns
    -------
    results: np.ndarray
        Results of the Monte Carlo simulation as a numpy array.

    Notes
    -----
    The engine executes parallel computing for number of simulation run larger than 1_000.
    Instead, serial computing is adopted.
    """
    # Generate project instances to be run in the simulation
    args = ["psc", "psc_arguments", "summary_arguments"]
    params = {arg: [0] * uncertainty_data["run_number"] for arg in args}

    for n in range(uncertainty_data["run_number"]):
        (
            params["psc"][n],
            params["psc_arguments"][n],
            params["summary_arguments"][n],
        ) = AdjustData(
            data=data,
            workbook_path=workbook_path,
            multipliers=mult[n, :],
        ).activate()

    # Execute MonteCarlo simulation
    results = np.zeros([uncertainty_data["run_number"], len(target)], dtype=np.float_)

    # Run serial for number of simulation less than 1000
    if uncertainty_data["run_number"] <= threshold_parallel:
        for n in range(uncertainty_data["run_number"]):
            results[n, :] = run_montecarlo(
                contract=params["psc"][n],
                contract_arguments=params["psc_arguments"][n],
                summary_arguments=params["summary_arguments"][n],
            )

    # Run parallel for number of simulation larger than 1000
    else:
        combined_iters = []
        for n in range(uncertainty_data["run_number"]):
            combined_iters.append(
                (
                    params["psc"][n],
                    params["psc_arguments"][n],
                    params["summary_arguments"][n],
                )
            )

        pl = mp.Pool(processes=int(os.cpu_count() - 1))
        results_init = pl.starmap(run_montecarlo, combined_iters)
        pl.close()
        pl.join()

        # Arrange the results into the appointed data structure
        for n in range(uncertainty_data["run_number"]):
            results[n, :] = results_init[n]

    return results



import numpy as np
from scipy.optimize import minimize_scalar

from pyscnomics.contracts.costrecovery import CostRecovery
from pyscnomics.contracts.grossplit import GrossSplit
from pyscnomics.contracts.transition import Transition
from pyscnomics.econ.selection import OptimizationParameter, OptimizationTarget, FluidType
from pyscnomics.tools.summary import get_summary

from pyscnomics.econ.costs import CapitalCost, Intangible, OPEX, ASR
from pyscnomics.optimize.optimization import adjust_useful_life_years


class OptimizationExceptionTransition(Exception):
    """Exception to raise for a misuse of BaseProject class"""

    pass


def adjust_contract(
        contract: Transition,
        contract_arguments: dict,
        variable: OptimizationParameter,
        value: float,
        summary_argument: dict,
        target_parameter: str
) -> (Transition, dict):
    """
    The function used to adjust the variable within a psc contract object.
    This function will be used by optimize_psc().

    Parameters
    ----------
    contract: CostRecovery | GrossSplit
        The contract object.
    contract_arguments: dict
        The contract arguments that passed on the fly .run() of the contract dataclass.
    variable: OptimizationParameter
        The enum selection of variable that will be changed.
    value: float
        The new value of the variable that will be replaced.
    summary_argument: dict
        The dictionary containing the arguments that passed summary function.
    target_parameter: str
        The string of the targeted parameter.

    Returns
    -------
    out: tuple
        The result of the target parameter and contract object that has been modified and run.

    """
    # When the optimization is VAT
    if variable is OptimizationParameter.VAT_DISCOUNT:
        contract = adjust_cost_element(contract=contract,
                                       adjustment_value=value,
                                       adjustment_variable=OptimizationParameter.VAT_DISCOUNT, )

    # When the optimization is LBT
    elif variable is OptimizationParameter.LBT_DISCOUNT:
        contract = adjust_cost_element(contract=contract,
                                       adjustment_value=value,
                                       adjustment_variable=OptimizationParameter.LBT_DISCOUNT, )

    elif variable is OptimizationParameter.DEPRECIATION_ACCELERATION:
        contract = adjust_cost_element(contract=contract,
                                       adjustment_value=value,
                                       adjustment_variable=OptimizationParameter.DEPRECIATION_ACCELERATION, )

    # The condition when contract is Cost Recovery
    if isinstance(contract.contract2, CostRecovery):
        # Changing the attributes of the contract based on the chosen variable
        if variable is OptimizationParameter.OIL_CTR_PRETAX:
            contract.contract2.oil_ctr_pretax_share = value

        if variable is OptimizationParameter.GAS_CTR_PRETAX:
            contract.contract2.gas_ctr_pretax_share = value

        if variable is OptimizationParameter.OIL_FTP_PORTION:
            contract.contract2.oil_ftp_portion = value

        if variable is OptimizationParameter.GAS_FTP_PORTION:
            contract.contract2.gas_ftp_portion = value

        if variable is OptimizationParameter.OIL_IC:
            contract.contract2.oil_ic_rate = value

        if variable is OptimizationParameter.GAS_IC:
            contract.contract2.gas_ic_rate = value

        if variable is OptimizationParameter.OIL_DMO_FEE:
            contract.contract2.oil_dmo_fee_portion = value

        if variable is OptimizationParameter.GAS_DMO_FEE:
            contract.contract2.gas_dmo_fee_portion = value

        if variable is OptimizationParameter.VAT_RATE:
            contract.argument_contract2['vat_rate'] = value

        if variable is OptimizationParameter.EFFECTIVE_TAX_RATE:
            contract.argument_contract2['tax_rate'] = value

    # The condition when contract is Gross Split
    if isinstance(contract.contract2, GrossSplit):
        # Changing the attributes of the contract based on the chosen variable
        if variable is OptimizationParameter.MINISTERIAL_DISCRETION:
            contract.contract2.split_ministry_disc = value

        if variable is OptimizationParameter.VAT_RATE:
            contract.argument_contract2['vat_rate'] = value

        if variable is OptimizationParameter.EFFECTIVE_TAX_RATE:
            contract.argument_contract2['tax_rate'] = value

    # Running the contract
    contract.run(**contract_arguments)

    # Get the summary of the new contract and get its value of the targeted optimization
    summary_argument['contract'] = contract
    result_psc = get_summary(**summary_argument)[target_parameter]

    return result_psc, contract


def adjust_cost_element(
        contract: Transition,
        adjustment_value: float = 1,
        adjustment_variable: OptimizationParameter = OptimizationParameter.VAT_DISCOUNT) -> Transition:
    """
    Function to adjust the vat discount or lbt discount of the cost element
    by multiplying them with the adjustment factor.

    Parameters
    ----------
    contract: Transition
        The contract which will be adjusted
    adjustment_value: float
        The factor which will be applied for the adjustment
    adjustment_variable: str
        The cost element that will be adjusted. The options are: "VAT" or "LBT"

    Returns
    -------
    contract_adjusted: Transition
        The contract that the cost element has been adjusted.

    """

    # Condition when the VAT of each cost element will be adjusted
    if adjustment_variable == OptimizationParameter.VAT_DISCOUNT:
        # Adjusting the Tangible cost of the contract
        tangible_adjusted = tuple([
            CapitalCost(start_year=tan.start_year,
                        end_year=tan.end_year,
                        cost=tan.cost,
                        expense_year=tan.expense_year,
                        cost_allocation=tan.cost_allocation,
                        description=tan.description,
                        vat_portion=tan.vat_portion,
                        vat_discount=adjustment_value,
                        lbt_portion=tan.lbt_portion,
                        lbt_discount=tan.lbt_discount,
                        pis_year=tan.pis_year,
                        salvage_value=tan.salvage_value,
                        useful_life=tan.useful_life,
                        depreciation_factor=tan.depreciation_factor,
                        is_ic_applied=tan.is_ic_applied,
                        ) for tan in contract.contract2.capital_cost
        ])

        # Adjusting the Intangible cost of the contract
        intangible_adjusted = tuple([
            Intangible(start_year=intang.start_year,
                       end_year=intang.end_year,
                       cost=intang.cost,
                       expense_year=intang.expense_year,
                       cost_allocation=intang.cost_allocation,
                       description=intang.description,
                       vat_portion=intang.vat_portion,
                       vat_discount=adjustment_value,
                       lbt_portion=intang.lbt_portion,
                       lbt_discount=intang.lbt_discount, ) for intang in contract.contract2.intangible_cost
        ])

        # Adjusting the OPEX cost of the contract
        opex_adjusted = tuple([
            OPEX(start_year=opx.start_year,
                 end_year=opx.end_year,
                 expense_year=opx.expense_year,
                 cost_allocation=opx.cost_allocation,
                 description=opx.description,
                 vat_portion=opx.vat_portion,
                 vat_discount=adjustment_value,
                 lbt_portion=opx.lbt_portion,
                 lbt_discount=opx.lbt_discount,
                 fixed_cost=opx.fixed_cost,
                 prod_rate=opx.prod_rate,
                 cost_per_volume=opx.cost_per_volume, ) for opx in contract.contract2.opex
        ])

        # Adjusting the ASR cost of the contract
        asr_adjusted = tuple([
            ASR(start_year=asr.start_year,
                end_year=asr.end_year,
                cost=asr.cost,
                expense_year=asr.expense_year,
                cost_allocation=asr.cost_allocation,
                description=asr.description,
                vat_portion=asr.vat_portion,
                vat_discount=adjustment_value,
                lbt_portion=asr.lbt_portion,
                lbt_discount=asr.lbt_discount, ) for asr in contract.contract2.asr_cost
        ])

    elif adjustment_variable == OptimizationParameter.LBT_DISCOUNT:
        # Adjusting the Tangible cost of the contract
        tangible_adjusted = tuple([
            CapitalCost(start_year=tan.start_year,
                        end_year=tan.end_year,
                        cost=tan.cost,
                        expense_year=tan.expense_year,
                        cost_allocation=tan.cost_allocation,
                        description=tan.description,
                        vat_portion=tan.vat_portion,
                        vat_discount=tan.vat_discount,
                        lbt_portion=tan.lbt_portion,
                        lbt_discount=adjustment_value,
                        pis_year=tan.pis_year,
                        salvage_value=tan.salvage_value,
                        useful_life=tan.useful_life,
                        depreciation_factor=tan.depreciation_factor,
                        is_ic_applied=tan.is_ic_applied,
                        ) for tan in contract.contract2.capital_cost
        ])

        # Adjusting the Intangible cost of the contract
        intangible_adjusted = tuple([
            Intangible(start_year=intang.start_year,
                       end_year=intang.end_year,
                       cost=intang.cost,
                       expense_year=intang.expense_year,
                       cost_allocation=intang.cost_allocation,
                       description=intang.description,
                       vat_portion=intang.vat_portion,
                       vat_discount=intang.vat_discount,
                       lbt_portion=intang.lbt_portion,
                       lbt_discount=adjustment_value, ) for intang in contract.contract2.intangible_cost
        ])

        # Adjusting the OPEX cost of the contract
        opex_adjusted = tuple([
            OPEX(start_year=opx.start_year,
                 end_year=opx.end_year,
                 expense_year=opx.expense_year,
                 cost_allocation=opx.cost_allocation,
                 description=opx.description,
                 vat_portion=opx.vat_portion,
                 vat_discount=opx.vat_discount,
                 lbt_portion=opx.lbt_portion,
                 lbt_discount=adjustment_value,
                 fixed_cost=opx.fixed_cost,
                 prod_rate=opx.prod_rate,
                 cost_per_volume=opx.cost_per_volume, ) for opx in contract.contract2.opex
        ])

        # Adjusting the ASR cost of the contract
        asr_adjusted = tuple([
            ASR(start_year=asr.start_year,
                end_year=asr.end_year,
                cost=asr.cost,
                expense_year=asr.expense_year,
                cost_allocation=asr.cost_allocation,
                description=asr.description,
                vat_portion=asr.vat_portion,
                vat_discount=asr.vat_discount,
                lbt_portion=asr.lbt_portion,
                lbt_discount=adjustment_value, ) for asr in contract.contract2.asr_cost
        ])

    elif adjustment_variable == OptimizationParameter.DEPRECIATION_ACCELERATION:
        # Adjusting the useful life of the capital cost of the contract
        tangible_adjusted = tuple([
            CapitalCost(start_year=tan.start_year,
                        end_year=tan.end_year,
                        cost=tan.cost,
                        expense_year=tan.expense_year,
                        cost_allocation=tan.cost_allocation,
                        description=tan.description,
                        vat_portion=tan.vat_portion,
                        vat_discount=tan.vat_discount,
                        lbt_portion=tan.lbt_portion,
                        lbt_discount=tan.lbt_discount,
                        pis_year=tan.pis_year,
                        salvage_value=tan.salvage_value,
                        useful_life=adjust_useful_life_years(
                            adjustment_value=adjustment_value,
                            useful_life_array=tan.useful_life),
                        depreciation_factor=tan.depreciation_factor,
                        is_ic_applied=tan.is_ic_applied,
                     ) for tan in contract.contract2.capital_cost
        ])
        intangible_adjusted = contract.contract2.intangible_cost
        opex_adjusted = contract.contract2.opex
        asr_adjusted = contract.contract2.asr_cost

    # Condition when the chosen option is not recognized
    else:
        raise OptimizationExceptionTransition(f"Adjustment Variable {adjustment_variable} "
                                              f"do not exist. It should be VAT or LBT in string data type")

    # Condition where only one fluid is produced
    produced_fluid = [i.fluid_type for i in contract.contract2.lifting]

    if FluidType.OIL not in produced_fluid:
        oil_onstream_date = None
    else:
        oil_onstream_date = contract.contract2.oil_onstream_date

    if FluidType.GAS not in produced_fluid:
        gas_onstream_date = None
    else:
        gas_onstream_date = contract.contract2.gas_onstream_date

    # When the contract is CostRecovery, parsing back the adjusted cost elements to the cost recovery contract
    if isinstance(contract.contract2, CostRecovery):
        contract2_adjusted = CostRecovery(start_date=contract.contract2.start_date,
                                          end_date=contract.contract2.end_date,
                                          oil_onstream_date=oil_onstream_date,
                                          gas_onstream_date=gas_onstream_date,
                                          lifting=contract.contract2.lifting,
                                          capital_cost=tangible_adjusted,
                                          intangible_cost=intangible_adjusted,
                                          opex=opex_adjusted,
                                          asr_cost=asr_adjusted,
                                          oil_ftp_is_available=contract.contract2.oil_ftp_is_available,
                                          oil_ftp_is_shared=contract.contract2.oil_ftp_is_shared,
                                          oil_ftp_portion=contract.contract2.oil_ftp_portion,
                                          gas_ftp_is_available=contract.contract2.gas_ftp_is_available,
                                          gas_ftp_is_shared=contract.contract2.gas_ftp_is_shared,
                                          gas_ftp_portion=contract.contract2.gas_ftp_portion,
                                          tax_split_type=contract.contract2.tax_split_type,
                                          condition_dict=contract.contract2.condition_dict,
                                          indicator_rc_icp_sliding=contract.contract2.indicator_rc_icp_sliding,
                                          oil_ctr_pretax_share=contract.contract2.oil_ctr_pretax_share,
                                          gas_ctr_pretax_share=contract.contract2.gas_ctr_pretax_share,
                                          oil_ic_rate=contract.contract2.oil_ic_rate,
                                          gas_ic_rate=contract.contract2.gas_ic_rate,
                                          ic_is_available=contract.contract2.ic_is_available,
                                          oil_cr_cap_rate=contract.contract2.oil_cr_cap_rate,
                                          gas_cr_cap_rate=contract.contract2.gas_cr_cap_rate,
                                          oil_dmo_volume_portion=contract.contract2.oil_dmo_volume_portion,
                                          oil_dmo_fee_portion=contract.contract2.oil_dmo_fee_portion,
                                          oil_dmo_holiday_duration=contract.contract2.oil_dmo_holiday_duration,
                                          gas_dmo_volume_portion=contract.contract2.gas_dmo_volume_portion,
                                          gas_dmo_fee_portion=contract.contract2.gas_dmo_fee_portion,
                                          gas_dmo_holiday_duration=contract.contract2.gas_dmo_holiday_duration, )

    # When the contract is GrossSplit, parsing back the adjusted cost elements to the gross split contract
    elif isinstance(contract.contract2, GrossSplit):
        contract2_adjusted = GrossSplit(start_date=contract.contract2.start_date,
                                        end_date=contract.contract2.end_date,
                                        oil_onstream_date=oil_onstream_date,
                                        gas_onstream_date=gas_onstream_date,
                                        lifting=contract.contract2.lifting,
                                        capital_cost=tangible_adjusted,
                                        intangible_cost=intangible_adjusted,
                                        opex=opex_adjusted,
                                        asr_cost=asr_adjusted,
                                        field_status=contract.contract2.field_status,
                                        field_loc=contract.contract2.field_loc,
                                        res_depth=contract.contract2.res_depth,
                                        infra_avail=contract.contract2.infra_avail,
                                        res_type=contract.contract2.res_type,
                                        api_oil=contract.contract2.api_oil,
                                        domestic_use=contract.contract2.domestic_use,
                                        prod_stage=contract.contract2.prod_stage,
                                        co2_content=contract.contract2.co2_content,
                                        h2s_content=contract.contract2.h2s_content,
                                        base_split_ctr_oil=contract.contract2.base_split_ctr_oil,
                                        base_split_ctr_gas=contract.contract2.base_split_ctr_gas,
                                        split_ministry_disc=contract.contract2.split_ministry_disc,
                                        oil_dmo_volume_portion=contract.contract2.oil_dmo_volume_portion,
                                        oil_dmo_fee_portion=contract.contract2.oil_dmo_fee_portion,
                                        oil_dmo_holiday_duration=contract.contract2.oil_dmo_holiday_duration,
                                        gas_dmo_volume_portion=contract.contract2.gas_dmo_volume_portion,
                                        gas_dmo_fee_portion=contract.contract2.gas_dmo_fee_portion,
                                        gas_dmo_holiday_duration=contract.contract2.gas_dmo_holiday_duration, )

    # When the contract is not recognized, raise an exception
    else:
        raise OptimizationExceptionTransition(f"Contract Type {type(contract)} , is not recognized for optimization module")

    # Replacing the second contract with the contract_adjusted
    contract_adjusted = contract
    contract_adjusted.contract2 = contract2_adjusted

    return contract_adjusted


def optimize_psc_core(
    dict_optimization: dict,
    contract: Transition,
    contract_arguments: dict,
    target_optimization_value: float,
    summary_argument: dict,
    target_parameter: OptimizationTarget = OptimizationTarget.IRR,
 ) -> (list, list, float, list):
    """
    The function to get contract variable(s) that resulting the desired target or contract's economic target.

    Parameters
    ----------
    dict_optimization: dict
        The optimization dictionary that containing the information about minimum boundary and upper boundary
        of the optimized parameters.
    contract: Transition
        The contract object.
    contract_arguments: dict
        The contract arguments that passed on the fly .run() of the contract dataclass.
    target_optimization_value: float
        The desired target value.
    summary_argument: dict
        The dictionary containing the arguments that passed summary function.
    target_parameter: OptimizationTarget
        The enum selection for economic indicator  that will be optimized.

    Notes
    -------
    The dictionary of dict_optimization should be at least having the structure as the following:
        dict_optimization = {'parameters': list[OptimizationParameter],
                             'min': np.ndarray,
                             'max': np.ndarray}

        'parameters' keys containing the enum list of the variable that will be optimized to achieve the target.
        'min' keys containing the minimum value of each parameter.
        'max' keys containing the minimum value of each parameter.


    Returns
    -------
    out : tuple

    list_str: list
        The list of parameter that passed and has been optimized.
    list_params_value: list
        The list of parameter's value that passed and has been optimized.
    result_optimization: float
        The value of the targeted parameter which the result of the optimization.
    list_executed_contract: list
        The list of executed contracts.
    """
    # Changing the Optimization selection from Enum to string in order to retrieve
    # the result from summary dictionary
    if target_parameter is OptimizationTarget.IRR:
        target_parameter = 'ctr_irr'

    elif target_parameter is OptimizationTarget.NPV:
        target_parameter = 'ctr_npv'

    elif target_parameter is OptimizationTarget.PI:
        target_parameter = 'ctr_pi'

    else:
        raise OptimizationExceptionTransition(
            f"target {target_parameter} "
            f"is should be one of: {[OptimizationTarget.value]}"
        )

    # Changing the parameters list[str] into list[OptimizationParameters(Enum)]
    list_params = dict_optimization['parameter']

    # Defining Base Value list to contain value of optimized parameters and status of the optimization
    list_params_value = ['Base Value'] * len(list_params)

    # Defining the empty result of optimization target, will be defined later
    result_optimization = None

    # Defining the executed contract list
    list_executed_contract = []

    psc = contract
    for index, param in enumerate(list_params):

        # Get the maximum value of each params
        if param is OptimizationParameter.OIL_CTR_PRETAX:
            max_value = dict_optimization['max'][index]
        elif param is OptimizationParameter.GAS_CTR_PRETAX:
            max_value = dict_optimization['max'][index]
        elif param is OptimizationParameter.OIL_FTP_PORTION:
            max_value = dict_optimization['min'][index]
        elif param is OptimizationParameter.GAS_FTP_PORTION:
            max_value = dict_optimization['min'][index]
        elif param is OptimizationParameter.OIL_IC:
            max_value = dict_optimization['max'][index]
        elif param is OptimizationParameter.GAS_IC:
            max_value = dict_optimization['max'][index]
        elif param is OptimizationParameter.OIL_DMO_FEE:
            max_value = dict_optimization['max'][index]
        elif param is OptimizationParameter.GAS_DMO_FEE:
            max_value = dict_optimization['max'][index]
        elif param is OptimizationParameter.VAT_RATE:
            max_value = dict_optimization['min'][index]
        elif param is OptimizationParameter.EFFECTIVE_TAX_RATE:
            max_value = dict_optimization['min'][index]
        elif param is OptimizationParameter.VAT_DISCOUNT:
            max_value = dict_optimization['max'][index]
        elif param is OptimizationParameter.LBT_DISCOUNT:
            max_value = dict_optimization['max'][index]
        elif param is OptimizationParameter.DEPRECIATION_ACCELERATION:
            max_value = dict_optimization['max'][index]
        elif param is OptimizationParameter.MINISTERIAL_DISCRETION:
            max_value = dict_optimization['max'][index]
        else:
            raise OptimizationExceptionTransition(f" Optimization parameter, {param}, is not recognized."
                                                  f" Optimization parameter should be chosen "
                                                  f"from OptimizationParameter enum.")

        # Changing the Contract parameter value based on the given input
        result_psc, psc = adjust_contract(contract=psc,
                                          contract_arguments=contract_arguments,
                                          variable=param,
                                          value=max_value,
                                          summary_argument=summary_argument,
                                          target_parameter=target_parameter, )

        # Able to conduct optimization since the result is greater than the target
        if result_psc > target_optimization_value:
            # Defining the upper and lower limit of the optimized variable
            bounds = (dict_optimization['min'][index],
                      dict_optimization['max'][index])

            def objective_run(new_value):
                result_psc_obj, executed_contract = adjust_contract(contract=psc,
                                                                    contract_arguments=contract_arguments,
                                                                    variable=param,
                                                                    value=new_value,
                                                                    summary_argument=summary_argument,
                                                                    target_parameter=target_parameter)

                result_obj = abs(result_psc_obj - target_optimization_value)
                return result_obj

            # Optimization of the objective function
            optim_result = minimize_scalar(objective_run, bounds=bounds, method='bounded')

            # Difference value from target and optimization result
            optimized_diff = optim_result.fun

            # Optimized Parameter Value
            optimized_parameter = optim_result.x

            # Result of the objective function
            function_result = optimized_diff + target_optimization_value

            # Writing the result of optimization to the list_params_value
            list_params_value[index] = optimized_parameter

            # Defining the result_optimization
            result_optimization = function_result

            # Defining the executed contract
            executed_contract = adjust_contract(contract=psc,
                                                contract_arguments=contract_arguments,
                                                variable=param,
                                                value=optimized_parameter,
                                                summary_argument=summary_argument,
                                                target_parameter=target_parameter)[1]

            # Filling the list with executed contract
            list_executed_contract.append(executed_contract)

            # # Printing for debugging
            # print('Parameter:', param)
            # print('Optimized Parameter Value:', optimized_parameter)
            # print('Targeted Value:', target_optimization_value)
            # print('Optimization Value:', function_result)
            # print('Difference of Target and Optimization Result:', optimized_diff)

            # Exiting the loop since the target has been achieved
            break

        elif result_psc <= target_optimization_value:
            # Writing the maximum value to the list_params_value
            list_params_value[index] = max_value

            # Defining the result_optimization
            result_optimization = result_psc

            list_executed_contract.append(psc)

    # Converting the list of enum into list of str enum value
    list_str = [enum_value.value for enum_value in list_params]

    # # Printing for debugging
    # print(list_str)
    # print(list_params_value)
    # print(result_optimization)

    return list_str, list_params_value, result_optimization, list_executed_contract



from dataclasses import dataclass, field
from typing import Union, Tuple
import numpy as np

from pyscnomics.contracts.project import BaseProject
from pyscnomics.contracts.costrecovery import CostRecovery
from pyscnomics.contracts.grossplit import GrossSplit
from pyscnomics.contracts.transition import Transition

from pyscnomics.econ.selection import NPVSelection, DiscountingMode

from pyscnomics.econ.indicator import (irr,
                                       npv_nominal_terms,
                                       npv_real_terms,
                                       npv_skk_nominal_terms,
                                       npv_skk_real_terms,
                                       npv_point_forward,
                                       pot_psc)


class CashflowException(Exception):
    """Exception to raise for a misuse of Cashflow Exception"""

    pass


@dataclass
class Cashflow:
    """
    A dataclass used to represent the cashflow for a project.

    Attributes
    ----------
    contract : (CostRecovery | GrossSplit | Transition | Tuple[Union[CostRecovery, GrossSplit, Transition], ...])
        The type of contract or a tuple of different contract types.
    reference_year : int, optional
        The reference year for the economic indicator (default is None).
    inflation_rate : float, optional
        The inflation rate applied to calculate the economic indicator (default is 0.0).
    discount_rate : float, optional
        The discount rate used for calculating NPV (default is 0.1).
    npv_mode : NPVSelection, optional
        The mode for calculating NPV, default is NPVSelection.NPV_SKK_REAL_TERMS.
    discounting_mode : DiscountingMode, optional
        The mode for discounting, default is DiscountingMode.END_YEAR.
    start_year : int, optional
        The start year of the cashflow, defined later (default is None, not included in initialization or representation).
    end_year : int, optional
        The end year of the cashflow, defined later (default is None, not included in initialization or representation).
    project_years : np.ndarray, optional
        The years of the project, defined later (default is None, not included in initialization or representation).
    cashflow : np.ndarray, optional
        The cashflow values for the project, defined later (default is None, not included in initialization or representation).
    npv : float, optional
        The Net Present Value (NPV) of the cashflow, defined later (default is None, not included in initialization or representation).
    irr : float, optional
        The Internal Rate of Return (IRR) of the cashflow, defined later (default is None, not included in initialization or representation).
    pot : float, optional
        The Payback on Time (POT) of the cashflow, defined later (default is None, not included in initialization or representation).
    """
    contract: (CostRecovery | GrossSplit | Transition |
               Tuple[Union[CostRecovery, GrossSplit, Transition], ...])

    reference_year: int = field(default=None)
    inflation_rate: float = field(default=0.0)
    discount_rate: float = field(default=0.1)
    npv_mode: NPVSelection = field(default=NPVSelection.NPV_SKK_REAL_TERMS)
    discounting_mode: DiscountingMode = field(default=DiscountingMode.END_YEAR)

    # Attributes associated with the date of the cashflow
    start_year: int = field(default=None, init=False, repr=False)
    end_year: int = field(default=None, init=False, repr=False)

    # Attributes associated with the years which will be defined later
    project_years: np.ndarray = field(default=None, init=False, repr=False)

    # Attributes associated with the economic indicator which will be defined later
    cashflow: np.ndarray = field(default=None, init=False, repr=False)
    npv: float = field(default=None, init=False, repr=False)
    irr: float = field(default=None, init=False, repr=False)
    pot: float = field(default=None, init=False, repr=False)

    def __post_init__(self):
        # Check if the contract/s has been run
        self.check_run_status()

        # Filling the start_year, end_year, and project_years
        if isinstance(self.contract, (BaseProject, CostRecovery, GrossSplit, Transition)):
            self.get_single_project_start_end()
            self.get_single_project_years()

        elif isinstance(self.contract, tuple):
            self.get_multiple_project_start_end()
            self.get_multiple_project_years()

        else:
            raise CashflowException(
                f"The contract type: {type(self.contract)} , is not recognized."
            )

        # Check the reference year
        if self.reference_year is None:
            # Retrieving the first year for the contract's project year as the reference year
            self.reference_year = min(self.project_years)
        else:
            self.check_reference_year()

    def check_run_status(self):
        """
        Function to check whether the contract has been run or not.
        """
        # Condition when the contract is a single-value
        if isinstance(self.contract, (BaseProject, CostRecovery, GrossSplit, Transition)):
            if self.contract._consolidated_cashflow is None:
                raise CashflowException(
                    f"Cashflow dataclass only receive contract that has been run. "
                    f"Please run the contract before parse it to the Cashflow dataclass."
                )
            else:
                pass

        # Condition when the contract is a multi-value
        elif isinstance(self.contract, tuple):
            for index, cntr in enumerate(self.contract):
                if cntr._consolidated_cashflow is None:
                    raise CashflowException(
                        f"Cashflow dataclass only receive contracts that has been run. "
                        f"Contract in index {index} has not been run, please run the contract before parse it to the "
                        f"Cashflow dataclass"
                    )
                else:
                    pass
        else:
            raise CashflowException(
                f"Cashflow dataclass only receive psc contract/s."
            )

    def get_single_project_start_end(self):
        """
        Function to parse the self.start_year and self.end_year
        """
        if self.start_year is None:
            self.start_year = self.contract.start_date.year
        else:
            pass

        if self.end_year is None:
            self.end_year = self.contract.end_date.year
        else:
            pass

    def get_multiple_project_start_end(self):
        """
        Function to parse the self.start_year and self.end_year from the given contracts.
        """
        self.start_year = min(np.array([cntr.start_date.year for cntr in self.contract], dtype=int))
        self.end_year = max(np.array([cntr.end_date.year for cntr in self.contract], dtype=int))

    def get_single_project_years(self):
        """
        Function to parse the self.start_year and self.end_year from the given contract.
        """
        self.project_years = self.contract.project_years

    def get_multiple_project_years(self):
        """
        Function to parse the self.project_years from the given contracts.
        """
        start_year_arr = min(np.array([cntr.start_date.year for cntr in self.contract], dtype=int))
        end_year_arr = max(np.array([cntr.end_date.year for cntr in self.contract], dtype=int))
        self.project_years = np.arange(start=start_year_arr, stop=end_year_arr, step=1)

    def check_reference_year(self):
        """
        Function to check the input of reference year.
        """
        if self.start_year is None:
            pass
        else:
            if self.reference_year < self.start_year:
                raise CashflowException(
                    f"The reference year: {self.reference_year}, is less than start_year {self.start_year}. "
                    f"The reference year should be in the range of the start_year and end_year"
                )
            elif self.reference_year > self.end_year:
                raise CashflowException(
                    f"The reference year: {self.reference_year}, is greater than end_year {self.end_year}. "
                    f"The reference year should be in the range of the start_year and end_year"
                )
            else:
                pass

    def get_cashflow_all(self):
        """
        Function to parse self.cashflow from the input of contract/s.
        """
        if isinstance(self.contract, (BaseProject, CostRecovery, GrossSplit, Transition)):
            self.cashflow = self.contract._consolidated_cashflow

        elif isinstance(self.contract, tuple):
            self.parse_multiple_cashflow()

        else:
            raise CashflowException(
                f"The contract type: {type(self.contract)} , is not recognized."
            )

    def parse_multiple_cashflow(self):
        """
        Function to retrieve the cashflow of the given contracts.
        """
        # Initiating the contract_years
        corresponding_year = self.project_years

        # Initiating the cashflow array
        self.cashflow = np.zeros_like(corresponding_year, dtype=float)

        # Iterate through each contract
        for cntr in self.contract:
            # Iterate through each year and cashflow in the contract
            for year, cashflow in zip(cntr.project_years, cntr._consolidated_cashflow):
                if year in corresponding_year:
                    # Find the index of the year in var_year
                    index = np.where(corresponding_year == year)[0][0]
                    # Add the cashflow to the total cashflows
                    self.cashflow[index] += cashflow

    def get_contract_npv(self):
        """
        Function to get the Net Present Value (NPV) of the obtained cashflow.
        """
        # NPV Calculation for SKK Real Terms
        if self.npv_mode == NPVSelection.NPV_SKK_REAL_TERMS:
            ctr_npv = npv_skk_real_terms(cashflow=self.cashflow,
                                         cashflow_years=self.project_years,
                                         discount_rate=self.discount_rate,
                                         reference_year=self.reference_year,
                                         discounting_mode=self.discounting_mode)

        # NPV Calculation for SKK Nominal Terms
        elif self.npv_mode == NPVSelection.NPV_SKK_NOMINAL_TERMS:
            ctr_npv = npv_skk_nominal_terms(cashflow=self.cashflow,
                                            cashflow_years=self.project_years,
                                            discount_rate=self.discount_rate,
                                            discounting_mode=self.discounting_mode)

        # NPV Calculation for Nominal Terms
        elif self.npv_mode == NPVSelection.NPV_NOMINAL_TERMS:
            ctr_npv = npv_nominal_terms(cashflow=self.cashflow,
                                        cashflow_years=self.project_years,
                                        discount_rate=self.discount_rate,
                                        reference_year=self.reference_year,
                                        discounting_mode=self.discounting_mode)

        # NPV Calculation for Real Terms
        elif self.npv_mode == NPVSelection.NPV_REAL_TERMS:
            ctr_npv = npv_real_terms(cashflow=self.cashflow,
                                     cashflow_years=self.project_years,
                                     discount_rate=self.discount_rate,
                                     reference_year=self.reference_year,
                                     inflation_rate=self.inflation_rate,
                                     discounting_mode=self.discounting_mode)

        # NPV Calculation for Point Forwards
        elif self.npv_mode == NPVSelection.NPV_POINT_FORWARD:
            ctr_npv = npv_point_forward(cashflow=self.cashflow,
                                        cashflow_years=self.project_years,
                                        discount_rate=self.discount_rate,
                                        reference_year=self.reference_year,
                                        discounting_mode=self.discounting_mode)

        # Condition when the npv_mode is not recognized
        else:
            raise CashflowException(
                f"The npv_mode: {self.npv_mode} , is not recognized."
                f"The npv_mode input should be in one of NPVSelection."
            )

        return ctr_npv

    def get_contract_irr(self):
        """
        Function to get the Internal Rate of Return (IRR) of the obtained cashflow.
        """
        return irr(cashflow=self.cashflow)

    def get_contract_pot(self):
        """
        Function to get the Pay-Out Time (POT) of the obtained cashflow.
        """
        return pot_psc(cashflow=self.cashflow,
                       cashflow_years=self.project_years,
                       reference_year=self.reference_year)

    def run(self):
        """
        Function to get the cashflow and economic indicator of the given contract/s.
        """
        # Retrieve the cashflow of the contract/s
        self.get_cashflow_all()

        # Retrieve the economic indicator
        self.npv = self.get_contract_npv()
        self.irr = self.get_contract_irr()
        self.pot = self.get_contract_pot()






"""
A unit testing for depreciation module.
"""

import numpy as np
from pyscnomics.dataset.sample import load_data, load_testing
from pyscnomics.econ.selection import InflationAppliedTo

psc = load_data(dataset_type='small_oil', contract_type='gross_split')
tax_rate = np.array([0.11, 0.11, 0.11, 0.12, 0.12, 0.12, 0.12, 0.12, 0.12, 0.12, 0.12, 0.12, 0.12, 0.12, 0.12, 0.12, 0.12, 0.12, 0.12, 0.12])
inflation_rate = np.array([0, 0, 0.0, 0.02, 0.02, 0.02, 0.02, 0.02, 0.02, 0.02, 0.02, 0.02, 0.02, 0.02, 0.02, 0.02, 0.02, 0.02, 0.02, 0.02])
psc.run(tax_rate=0.22,
        vat_rate=tax_rate,
        inflation_rate=inflation_rate,
        inflation_rate_applied_to=InflationAppliedTo.CAPEX)



def test_tangible():
    base = np.asarray(load_testing(dataset_type='small_oil', key='oil_tangible'))
    calc = psc._oil_capital_expenditures
    np.testing.assert_allclose(base, calc)


def test_intangible():
    base = np.asarray(load_testing(dataset_type='small_oil', key='oil_intangible'))
    calc = psc._oil_intangible_expenditures
    np.testing.assert_allclose(base, calc)


def test_opex():
    base = np.asarray(load_testing(dataset_type='small_oil', key='oil_opex'))
    calc = psc._oil_opex_expenditures
    np.testing.assert_allclose(base, calc)


def test_lifting():
    base = np.asarray(load_testing(dataset_type='small_oil', key='oil_lifting'))
    calc = psc._oil_lifting.get_lifting_rate_arr()
    np.testing.assert_allclose(base, calc)


def test_price():
    base = np.asarray(load_testing(dataset_type='small_oil', key='oil_price'))
    calc = psc._oil_lifting.get_price_arr()
    np.testing.assert_allclose(base, calc)


def test_revenue():
    base = np.asarray(load_testing(dataset_type='small_oil', key='oil_revenue'))
    calc = psc._oil_revenue
    np.testing.assert_allclose(base, calc)


def test_depreciation():
    base = np.asarray(load_testing(dataset_type='small_oil', key='oil_depreciation'))
    calc = np.array(psc._oil_depreciation)
    np.testing.assert_allclose(base, calc)


def test_non_capital():
    base = np.asarray(load_testing(dataset_type='small_oil', key='oil_non_capital'))
    calc = np.array(psc._oil_non_capital)
    np.testing.assert_allclose(base, calc)


def test_cumulative_prod():
    base = np.asarray(load_testing(dataset_type='small_oil', key='cumulative_prod'))
    calc = np.array(psc._cumulative_prod)
    np.testing.assert_allclose(base, calc)


def test_prog_split():
    base = np.asarray(load_testing(dataset_type='small_oil', key='oil_prog_split'))
    calc = np.array(psc._oil_prog_split)
    np.testing.assert_allclose(base, calc)


def test_oil_ctr_split():
    base = np.asarray(load_testing(dataset_type='small_oil', key='oil_ctr_split'))
    calc = np.array(psc._oil_ctr_split)
    np.testing.assert_allclose(base, calc)


def test_oil_ctr_share():
    base = np.asarray(load_testing(dataset_type='small_oil', key='oil_ctr_share'))
    calc = np.array(psc._oil_ctr_share_before_transfer)
    np.testing.assert_allclose(base, calc)


def test_oil_gov_share():
    base = np.asarray(load_testing(dataset_type='small_oil', key='oil_gov_share'))
    calc = np.array(psc._oil_gov_share)
    np.testing.assert_allclose(base, calc)


def test_oil_total_expense():
    base = np.asarray(load_testing(dataset_type='small_oil', key='oil_total_expenses'))
    calc = np.array(psc._oil_total_expenses)
    np.testing.assert_allclose(base, calc)


def test_oil_cost_tobe_deducted():
    base = np.asarray(load_testing(dataset_type='small_oil', key='oil_cost_tobe_deducted'))
    calc = np.array(psc._oil_cost_tobe_deducted)
    np.testing.assert_allclose(base, calc)


def test_oil_carry_forward_deductible_cost():
    base = np.asarray(load_testing(dataset_type='small_oil', key='oil_carward_deduct_cost'))
    calc = np.array(psc._oil_carward_deduct_cost)
    np.testing.assert_allclose(base, calc)


def test_oil_deductible_cost():
    base = np.asarray(load_testing(dataset_type='small_oil', key='oil_deductible_cost'))
    calc = np.array(psc._oil_deductible_cost)
    np.testing.assert_allclose(base, calc)


def test_oil_net_operating_profit():
    base = np.asarray(load_testing(dataset_type='small_oil', key='oil_net_operating_profit'))
    calc = np.array(psc._oil_net_operating_profit)
    np.testing.assert_allclose(base, calc)


def test_oil_dmo_volume():
    base = np.asarray(load_testing(dataset_type='small_oil', key='oil_dmo_volume'))
    calc = np.array(psc._oil_dmo_volume)
    np.testing.assert_allclose(base, calc)


def test_oil_dmo_fee():
    base = np.asarray(load_testing(dataset_type='small_oil', key='oil_dmo_fee'))
    calc = np.array(psc._oil_dmo_fee)
    np.testing.assert_allclose(base, calc)


def test_oil_ddmo():
    base = np.asarray(load_testing(dataset_type='small_oil', key='oil_ddmo'))
    calc = np.array(psc._oil_ddmo)
    np.testing.assert_allclose(base, calc)


def test_oil_taxable_income():
    base = np.asarray(load_testing(dataset_type='small_oil', key='oil_taxable_income'))
    calc = np.array(psc._oil_taxable_income)
    np.testing.assert_allclose(base, calc)


def test_oil_tax():
    base = np.asarray(load_testing(dataset_type='small_oil', key='oil_tax'))
    calc = np.array(psc._oil_tax)
    np.testing.assert_allclose(base, calc)


def test_oil_ctr_net_share():
    base = np.asarray(load_testing(dataset_type='small_oil', key='oil_ctr_net_share'))
    calc = np.array(psc._oil_ctr_net_share)
    np.testing.assert_allclose(base, calc)


def test_oil_ctr_cashflow():
    base = np.asarray(load_testing(dataset_type='small_oil', key='oil_ctr_cashflow'))
    calc = np.array(psc._oil_ctr_cashflow)
    np.testing.assert_allclose(base, calc)


def test_oil_gov_take():
    base = np.asarray(load_testing(dataset_type='small_oil', key='oil_gov_take'))
    calc = np.array(psc._oil_government_take)
    np.testing.assert_allclose(base, calc)


import numpy as np

from pyscnomics.econ.selection import DiscountingMode
from pyscnomics.econ.indicator import (npv_nominal_terms,
                                       npv_real_terms,
                                       npv_skk_nominal_terms,
                                       npv_skk_real_terms,
                                       npv_point_forward,
                                       pot_psc)

# Cashflow
cashflow_years = np.arange(2018, 2029)
cashflow = np.array([-100000, 0, 0, 0, 0, 0, 0, 50000, 50000, 50000, 50000])
discount_rate = 0.1
reference_year = 2023
inflation_rate = 0.03


def test_npv_nominal_terms():
    base = 44084.7930158769
    calc = npv_nominal_terms(cashflow=cashflow,
                             cashflow_years=cashflow_years,
                             discount_rate=discount_rate,
                             reference_year=reference_year,
                             discounting_mode=DiscountingMode.END_YEAR)
    np.testing.assert_allclose(base, calc)


def test_npv_real_terms():
    base = 28157.385585876900
    calc = npv_real_terms(cashflow=cashflow,
                          cashflow_years=cashflow_years,
                          discount_rate=discount_rate,
                          reference_year=reference_year,
                          inflation_rate=inflation_rate,
                          discounting_mode=DiscountingMode.END_YEAR)

    np.testing.assert_allclose(base, calc)


def test_npv_skk_nominal_terms():
    base = -10534.679687877200
    calc = npv_skk_nominal_terms(cashflow=cashflow,
                                 cashflow_years=cashflow_years,
                                 discount_rate=discount_rate,
                                 discounting_mode=DiscountingMode.END_YEAR)

    np.testing.assert_allclose(base, calc)


def test_npv_skk_real_terms():
    base = -16966.206984123100
    calc = npv_skk_real_terms(cashflow=cashflow,
                              cashflow_years=cashflow_years,
                              discount_rate=discount_rate,
                              reference_year=reference_year,
                              discounting_mode=DiscountingMode.END_YEAR)

    np.testing.assert_allclose(base, calc)


def test_npv_point_forward():
    base = 144084.793015877000
    calc = npv_point_forward(cashflow=cashflow,
                             cashflow_years=cashflow_years,
                             discount_rate=discount_rate,
                             reference_year=reference_year,
                             discounting_mode=DiscountingMode.END_YEAR)

    np.testing.assert_allclose(base, calc)


def test_pot():
    cashflow_psc = np.array(
        [-0.256683836, -0.673626192, -6.517022729, -0.386445703, -0.666439491, -8.706499027, -0.253577600, -0.128721715,
         -0.084977677, -0.572136045, -5.876060455, -1.430334000, -5.554723774, -36.792930707, 20.719872793,
         21.207345549, 21.069649586, 8.199755855, 5.989491246, 7.595050463, 7.223124711, 6.932400441, 3.556406469,
         6.819592533, 6.034993688, 5.108545699, 4.154169710, 2.793870918, 1.867523855, 1.224588341, 0.000000000,
         0.000000000, 0.000000000, 0.000000000, 0.000000000, 0.000000000])
    cashflow_years_pot = np.arange(2010, 2046)
    reference_year_pot = 2022

    base = 5.59798256
    calc = pot_psc(cashflow=cashflow_psc,
                   cashflow_years=cashflow_years_pot,
                   reference_year=reference_year_pot)

    np.testing.assert_allclose(base, calc)



import matplotlib.pyplot as plt
import matplotlib.ticker as mticker
import pandas as pd


def get_uncertainty_plot(uncertainty_outcomes: dict,
                         plot_type: str = 'Stairway'
                         ):
    """
    Function to get the uncertainty plot.

    Parameters
    ----------
    uncertainty_outcomes: dict
        The result of an uncertainty analyis in form of a dictionary.
    plot_type: str
        The type of the plot that will be generated. the available options is
        ['Stairway', 'Histogram']

    Returns
    -------
    Generating an uncertainty plot.

    """
    # Grouping the result
    result_freq = uncertainty_outcomes['results'][:, 0]
    result_npv = uncertainty_outcomes['results'][:, 1]
    result_irr = uncertainty_outcomes['results'][:, 2]
    result_pi = uncertainty_outcomes['results'][:, 3]
    result_pot = uncertainty_outcomes['results'][:, 4]
    result_gov_take = uncertainty_outcomes['results'][:, 5]
    result_ctr_net_share = uncertainty_outcomes['results'][:, 6]

    if plot_type == 'Stairway':
        # Create a figure and a grid of 2x3 subplots
        fig, axs = plt.subplots(2, 3, figsize=(18, 11))

        # plot each subplot individually
        # Configuring NPV
        axs[0, 0].set_title('NPV')
        axs[0, 0].plot(result_npv, result_freq, drawstyle="steps-post")
        axs[0, 0].set_xlabel("MUSD", fontsize=10, labelpad=0)
        axs[0, 0].set_ylabel("Frequency", fontsize=10, labelpad=0)

        locator_npv = mticker.MaxNLocator(nbins='auto')
        axs[0, 0].xaxis.set_major_locator(locator_npv)
        axs[0, 0].grid(which="major", linewidth=1)
        axs[0, 0].grid(which="minor", linewidth=0.2)
        axs[0, 0].minorticks_on()

        # IRR
        # Configuring IRR
        axs[0, 1].set_title('IRR')
        axs[0, 1].plot(result_irr, result_freq, drawstyle="steps-post")
        axs[0, 1].set_xlabel("Fraction", fontsize=10, labelpad=0)
        axs[0, 1].set_ylabel("Frequency", fontsize=14, labelpad=8)

        locator_irr = mticker.MaxNLocator(nbins='auto')
        axs[0, 1].xaxis.set_major_locator(locator_irr)

        axs[0, 1].grid(which="major", linewidth=1)
        axs[0, 1].grid(which="minor", linewidth=0.2)
        axs[0, 1].minorticks_on()

        # PI
        # Configuring PI
        axs[0, 2].set_title('PI')
        axs[0, 2].plot(result_pi, result_freq, drawstyle="steps-post")
        axs[0, 2].set_xlabel("Fraction", fontsize=10, labelpad=0)
        axs[0, 2].set_ylabel("Frequency", fontsize=14, labelpad=8)

        locator_pi = mticker.MaxNLocator(nbins='auto')
        axs[0, 2].xaxis.set_major_locator(locator_pi)

        axs[0, 2].grid(which="major", linewidth=1)
        axs[0, 2].grid(which="minor", linewidth=0.2)
        axs[0, 2].minorticks_on()

        # POT
        # Configuring POT
        axs[1, 0].set_title('POT')
        axs[1, 0].plot(result_pot, result_freq, drawstyle="steps-post")
        axs[1, 0].set_xlabel("Years", fontsize=10, labelpad=0)
        axs[1, 0].set_ylabel("Frequency", fontsize=14, labelpad=8)

        locator_pot = mticker.MaxNLocator(nbins='auto')
        axs[1, 0].xaxis.set_major_locator(locator_pot)

        axs[1, 0].grid(which="major", linewidth=1)
        axs[1, 0].grid(which="minor", linewidth=0.2)
        axs[1, 0].minorticks_on()

        # Government Take
        # Configuring Gov Take
        axs[1, 1].set_title('Gov. Take')
        axs[1, 1].plot(result_gov_take, result_freq, drawstyle="steps-post")
        axs[1, 1].set_xlabel("MUSD", fontsize=10, labelpad=0)
        axs[1, 1].set_ylabel("Frequency", fontsize=14, labelpad=8)

        locator_govtake = mticker.MaxNLocator('auto')
        axs[1, 1].xaxis.set_major_locator(locator_govtake)

        axs[1, 1].grid(which="major", linewidth=1)
        axs[1, 1].grid(which="minor", linewidth=0.2)
        axs[1, 1].minorticks_on()

        # Contractor Net Share
        # Configuring Gov Take
        axs[1, 2].set_title('CTR Net Share')
        axs[1, 2].plot(result_ctr_net_share, result_freq, drawstyle="steps-post")
        axs[1, 2].set_xlabel("MUSD", fontsize=10, labelpad=0)
        axs[1, 2].set_ylabel("Frequency", fontsize=14, labelpad=8)

        locator_netshare = mticker.MaxNLocator('auto')
        axs[1, 2].xaxis.set_major_locator(locator_netshare)

        axs[1, 2].grid(which="major", linewidth=1)
        axs[1, 2].grid(which="minor", linewidth=0.2)
        axs[1, 2].minorticks_on()

        # Adjust layout to prevent clipping of titles
        fig.subplots_adjust(wspace=10)
        plt.tight_layout(pad=5)
        plt.savefig('uncertainty_plot.png')
        # plt.show()

    elif plot_type == 'Histogram':
        df = pd.DataFrame()
        df['npv'] = result_npv
        df['irr'] = result_irr
        df['pi'] = result_pi
        df['pot'] = result_pot
        df['gov_take'] = result_gov_take
        df['ctr_net_share'] = result_ctr_net_share

        df.hist(figsize=(18, 10))
        plt.savefig('uncertainty_plot.png')
        # plt.show()






import numpy as np
from datetime import datetime

from pyscnomics.dataset.sample import load_data
from pyscnomics.econ.selection import FTPTaxRegime, FluidType, InflationAppliedTo
from pyscnomics.econ.revenue import Lifting
from pyscnomics.econ.costs import CapitalCost, Intangible, OPEX, ASR
from pyscnomics.contracts.costrecovery import CostRecovery
from pyscnomics.contracts.grossplit import GrossSplit
from pyscnomics.contracts.transition import Transition


def generate_contract_sample(case: str) -> CostRecovery | GrossSplit | Transition:
    if case == 'CR_1':
        # Initiating the contract
        psc = load_data(dataset_type='case1', contract_type='cost_recovery')
        psc.oil_ftp_is_available = True
        psc.oil_ftp_is_shared = True
        psc.oil_ftp_portion = 0.20
        psc.gas_ftp_is_available = True
        psc.gas_ftp_is_shared = True
        psc.gas_ftp_portion = 0.20

        # Split Pre Tax
        psc.oil_ctr_pretax_share = 0.34722220
        psc.gas_ctr_pretax_share = 0.5208330

        # DMO
        psc.oil_dmo_volume_portion = 0.25
        psc.oil_dmo_fee_portion = 0.25
        psc.oil_dmo_holiday_duration = 60
        psc.gas_dmo_volume_portion = 0.25
        psc.gas_dmo_fee_portion = 1
        psc.gas_dmo_holiday_duration = 60

        # Generating contract arguments
        contract_arguments = {'tax_rate': 0.424,
                              'ftp_tax_regime': FTPTaxRegime.PRE_PDJP_20_2017,
                              'sunk_cost_reference_year': 2021}

        psc.run(**contract_arguments)
        return psc

    elif case == 'CR_2':
        # Initiating the psc object
        psc = load_data(dataset_type='case2', contract_type='cost_recovery')

        # Editing the CostRecovery attribute as the corresponding case 1
        # FTP
        psc.oil_ftp_is_available = True
        psc.oil_ftp_is_shared = True
        psc.oil_ftp_portion = 0.2
        psc.gas_ftp_is_available = True
        psc.gas_ftp_is_shared = True
        psc.gas_ftp_portion = 0.05

        # Split Pre Tax
        psc.oil_ctr_pretax_share = 0.34722220
        psc.gas_ctr_pretax_share = 0.72083300

        # DMO
        psc.oil_dmo_volume_portion = 0.25
        psc.oil_dmo_fee_portion = 0.25
        psc.oil_dmo_holiday_duration = 60
        psc.gas_dmo_volume_portion = 0.25
        psc.gas_dmo_fee_portion = 1.0
        psc.gas_dmo_holiday_duration = 60

        # Generating contract arguments
        contract_arguments = {'tax_rate': 0.424,
                              'ftp_tax_regime': FTPTaxRegime.PRE_PDJP_20_2017,
                              'sunk_cost_reference_year': 2021}

        psc.run(**contract_arguments)
        return psc

    elif case == 'CR_3':
        # Initiating the psc object
        psc = load_data(dataset_type='case3', contract_type='cost_recovery')

        # Editing the CostRecovery attribute as the corresponding case 1
        # FTP
        psc.oil_ftp_is_available = True
        psc.oil_ftp_is_shared = True
        psc.oil_ftp_portion = 0.20
        psc.gas_ftp_is_available = True
        psc.gas_ftp_is_shared = True
        psc.gas_ftp_portion = 0.20

        # Split Pre Tax
        psc.oil_ctr_pretax_share = 0.34722220
        psc.gas_ctr_pretax_share = 0.5208330

        # DMO
        psc.oil_dmo_volume_portion = 0.25
        psc.oil_dmo_fee_portion = 0.25
        psc.oil_dmo_holiday_duration = 60
        psc.gas_dmo_volume_portion = 0.25
        psc.gas_dmo_fee_portion = 1
        psc.gas_dmo_holiday_duration = 60

        # Generating contract arguments
        contract_arguments = {'tax_rate': 0.424,
                              'ftp_tax_regime': FTPTaxRegime.PRE_PDJP_20_2017,
                              'sunk_cost_reference_year': 2021}

        psc.run(**contract_arguments)
        return psc

    elif case == 'GS_1':
        # Initiating the psc object
        psc = load_data(dataset_type='small_oil', contract_type='gross_split')
        tax_rate = np.array(
            [0.11, 0.11, 0.11, 0.12, 0.12, 0.12, 0.12, 0.12, 0.12, 0.12, 0.12, 0.12, 0.12, 0.12, 0.12, 0.12, 0.12, 0.12,
             0.12, 0.12])
        inflation_rate = np.array(
            [0, 0, 0.0, 0.02, 0.02, 0.02, 0.02, 0.02, 0.02, 0.02, 0.02, 0.02, 0.02, 0.02, 0.02, 0.02, 0.02, 0.02, 0.02,
             0.02])

        contract_arguments = {'tax_rate': 0.22,
                              'vat_rate': tax_rate,
                              'inflation_rate': inflation_rate,
                              'inflation_rate_applied_to': InflationAppliedTo.CAPEX}

        psc.run(**contract_arguments)
        return psc

    elif case == 'Transition_1':
        # Defining Start Date and End Date
        psc_1_start_date = datetime.strptime("01/01/2018", '%d/%m/%Y').date()
        psc_1_end_date = datetime.strptime("22/4/2020", '%d/%m/%Y').date()

        # Defining the Gas lifting data
        psc1_gas_lifting = Lifting(
            start_year=2018,
            end_year=2020,
            lifting_rate=np.array([1.58446786291200, 0.98961346458056]),
            price=np.array([6.260, 6.260]),
            prod_year=np.array([2019, 2020]),
            ghv=np.array([1047.0, 1047.0]),
            fluid_type=FluidType.GAS)

        # Defining the Gas Tangible Data - Drilling Tangible
        psc1_gas_tangible = CapitalCost(
            start_year=2018,
            end_year=2020,
            cost=np.array([3363.67743703704000, 802.73224043715800]),
            expense_year=np.array([2019, 2020]),
            cost_allocation=[FluidType.GAS] * 2,
            pis_year=np.array([2019, 2020]),
            useful_life=np.array([5, 5]),
            depreciation_factor=np.array([0.25, 0.25]),
        )

        # Defining the Gas Intangible Data
        psc1_gas_intang = Intangible(
            start_year=2018,
            end_year=2020,
            cost=np.array([9532.633600000]),
            expense_year=np.array([2019]),
            cost_allocation=[FluidType.GAS])

        # Defining the Gas OPEX Data
        psc1_gas_opex_cost = OPEX(
            start_year=2018,
            end_year=2020,
            fixed_cost=np.array([2076.908222642980, 1297.582047244550]),
            expense_year=np.array([2019, 2020]),
            cost_allocation=[FluidType.GAS] * 2)

        # Defining the Gas ASR Data
        psc1_gas_asr_cost_opx = OPEX(
            start_year=2018,
            end_year=2020,
            fixed_cost=np.array([35.515809523809900, 10.965263596148900]),
            expense_year=np.array([2019, 2020]),
            cost_allocation=[FluidType.GAS] * 2)

        psc1_gas_asr_cost = ASR(
            start_year=2018,
            end_year=2020,
            cost=np.array([0]),
            expense_year=np.array([2019]),
            cost_allocation=[FluidType.GAS])

        # Parsing the fiscal terms into Cost Recovery
        psc1 = CostRecovery(
            start_date=psc_1_start_date,
            end_date=psc_1_end_date,
            lifting=tuple([psc1_gas_lifting]),
            capital_cost=tuple([psc1_gas_tangible]),
            intangible_cost=tuple([psc1_gas_intang]),
            opex=tuple([psc1_gas_opex_cost, psc1_gas_asr_cost_opx]),
            asr_cost=tuple([psc1_gas_asr_cost]),
            oil_ftp_is_available=True,
            oil_ftp_is_shared=True,
            oil_ftp_portion=0.2,
            gas_ftp_is_available=True,
            gas_ftp_is_shared=True,
            gas_ftp_portion=0.2,
            oil_ctr_pretax_share=0.3361,
            gas_ctr_pretax_share=0.57692307692307700,
            oil_dmo_volume_portion=0.25,
            oil_dmo_fee_portion=0.15,
            oil_dmo_holiday_duration=0,
            gas_dmo_volume_portion=0.25,
            gas_dmo_fee_portion=0.15,
            gas_dmo_holiday_duration=0)

        """
        Gross Split
        """
        # Defining Start Date and End Date
        psc_2_start_date = datetime.strptime("23/4/2020", '%d/%m/%Y').date()
        psc_2_end_date = datetime.strptime("22/4/2030", '%d/%m/%Y').date()

        # Defining the Gas lifting data
        psc2_gas_lifting = Lifting(
            start_year=2020,
            end_year=2030,
            lifting_rate=np.array(
                [2.21568324370692000, 3.20769606628721000, 3.29284116326400000, 3.37370744832000000,
                 3.44718555313867000, 3.40400062841705000, 3.32543155814400000, 2.00043667046400000]),
            price=np.array([6.2600, 6.2600, 6.2600, 6.2600, 6.2600, 6.2600, 6.2600, 6.2600]),
            prod_year=np.array([2020, 2021, 2022, 2023, 2024, 2025, 2026, 2027]),
            ghv=np.array([1047.0, 1047.0, 1047.0, 1047.0, 1047.0, 1047.0, 1047.0, 1047.0, ]),
            fluid_type=FluidType.GAS)

        # Defining the Gas Tangible Data - Drilling Tangible
        psc2_gas_tangible = CapitalCost(
            start_year=2020,
            end_year=2030,
            cost=np.array([1959.561038251370, 2834.780000000000]),
            expense_year=np.array([2020, 2025]),
            pis_year=np.array([2020, 2025]),
            useful_life=np.array([5, 5]),
            depreciation_factor=np.array([0.25, 0.25]),
            cost_allocation=[FluidType.GAS] * 2)

        # Defining the Gas Intangible Data
        psc2_gas_intang = Intangible(
            start_year=2020,
            end_year=2030,
            cost=np.array([0]),
            expense_year=np.array([2020]),
            cost_allocation=[FluidType.GAS])

        # Defining the Gas OPEX Data
        psc2_gas_opex_cost = OPEX(
            start_year=2020,
            end_year=2030,
            fixed_cost=np.array(
                [3137.4561521272200, 3834.8277754882300, 3943.8595100171400, 4040.8953543109500, 4114.6574450911200,
                 4272.4109557000400, 5057.6546904997800, 4462.7942473087100, ]),
            expense_year=np.array([2020, 2021, 2022, 2023, 2024, 2025, 2026, 2027]),
            cost_allocation=[FluidType.GAS] * 8)

        psc2_gas_lbt_cost = OPEX(
            start_year=2020,
            end_year=2030,
            fixed_cost=np.array(
                [636.40338233533300, 1062.86476325927000, 1091.05051487026000, 1117.33210409363000, 1140.29666892703000,
                 1296.08012023605000, 1124.78850166335000, 786.84662704513200]),
            expense_year=np.array([2020, 2021, 2022, 2023, 2024, 2025, 2026, 2027]),
            cost_allocation=[FluidType.GAS] * 8)

        # Defining VAT and Import Duty
        psc2_gas_vat = OPEX(
            start_year=2020,
            end_year=2030,
            fixed_cost=np.array(
                [396.5062479576550, 286.7125283653920, 294.7836073052500, 301.9666899432990, 307.4269319547460,
                 553.8846425099840, 377.2323444857470, 333.1977740376320, ]),
            expense_year=np.array([2020, 2021, 2022, 2023, 2024, 2025, 2026, 2027]),
            cost_allocation=[FluidType.GAS] * 8)

        psc2_gas_import_duty = OPEX(
            start_year=2020,
            end_year=2030,
            fixed_cost=np.array(
                [159.965353, 55.922218, 57.496452, 58.897486, 59.962486, 227.566547, 73.577773, 64.988993]),
            expense_year=np.array([2020, 2021, 2022, 2023, 2024, 2025, 2026, 2027]),
            cost_allocation=[FluidType.GAS] * 8)

        # Defining the Gas ASR Data
        psc2_gas_asr_cost_opx = OPEX(
            start_year=2020,
            end_year=2030,
            fixed_cost=np.array(
                [26.513186, 38.355043, 38.355043, 38.355043, 38.355043, 38.355043, 38.355043, 38.355043, ]),
            expense_year=np.array([2020, 2021, 2022, 2023, 2024, 2025, 2026, 2027]),
            cost_allocation=[FluidType.GAS] * 8)

        psc2_gas_asr_cost = ASR(
            start_year=2020,
            end_year=2030,
            cost=np.array([0]),
            expense_year=np.array([2020]),
            cost_allocation=[FluidType.GAS])

        # Parsing the fiscal terms into Gross Split
        psc2 = GrossSplit(
            start_date=psc_2_start_date,
            end_date=psc_2_end_date,
            lifting=tuple([psc2_gas_lifting]),
            capital_cost=tuple([psc2_gas_tangible]),
            intangible_cost=tuple([psc2_gas_intang]),
            opex=tuple([psc2_gas_opex_cost, psc2_gas_asr_cost_opx, psc2_gas_lbt_cost]),
            asr_cost=tuple([psc2_gas_asr_cost]),
            field_status='No POD',
            field_loc='Onshore',
            res_depth='<=2500',
            infra_avail='Well Developed',
            res_type='Conventional',
            api_oil='>=25',
            domestic_use='70<=x<100',
            prod_stage='Primer',
            co2_content='<5',
            h2s_content='<100',
            base_split_ctr_oil=0.43,
            base_split_ctr_gas=0.48,
            split_ministry_disc=0.0,
            oil_dmo_volume_portion=0.25,
            oil_dmo_fee_portion=1.0,
            oil_dmo_holiday_duration=0,
            gas_dmo_volume_portion=0.25,
            gas_dmo_fee_portion=1.0,
            gas_dmo_holiday_duration=0)

        ftp_tax_regime = FTPTaxRegime.DIRECT_MODE
        eff_tax_rate = 0.48
        # tax_payment_method = TaxPaymentMode.TAX_DIRECT_MODE
        argument_contract1 = {'ftp_tax_regime': ftp_tax_regime,
                              'tax_rate': eff_tax_rate,
                              # 'tax_payment_method': tax_payment_method
                              }

        argument_contract2 = {'tax_rate': 0.40}

        psc_trans = Transition(contract1=psc1,
                               contract2=psc2,
                               argument_contract1=argument_contract1,
                               argument_contract2=argument_contract2)
        psc_trans.run(unrec_portion=0)
        return psc_trans
